@@ -23,6 +23,7 @@
 #include <config.h>
 #endif
 
+#include <errno.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
@@ -44,8 +45,9 @@ struct connman_route {
 };
 
 struct connman_provider {
-	struct connman_element element;
+	int refcount;
 	struct connman_service *vpn_service;
+	int index;
 	char *identifier;
 	char *name;
 	char *type;
@@ -55,6 +57,7 @@ struct connman_provider {
 	GHashTable *routes;
 	struct connman_provider_driver *driver;
 	void *driver_data;
+	GHashTable *setting_strings;
 };
 
 void __connman_provider_append_properties(struct connman_provider *provider,
@@ -73,6 +76,73 @@ void __connman_provider_append_properties(struct connman_provider *provider,
 						 &provider->type);
 }
 
+static int connman_provider_load(struct connman_provider *provider)
+{
+	gsize idx = 0;
+	GKeyFile *keyfile;
+	gchar **settings;
+	gchar *key, *value;
+	gsize length;
+
+	DBG("provider %p", provider);
+
+	keyfile = __connman_storage_load_provider(provider->identifier);
+	if (keyfile == NULL)
+		return -ENOENT;
+
+	settings = g_key_file_get_keys(keyfile, provider->identifier, &length,
+				NULL);
+	if (settings == NULL) {
+		g_key_file_free(keyfile);
+		return -ENOENT;
+	}
+
+	while (idx < length) {
+		key = settings[idx];
+		DBG("found key %s", key);
+		if (key != NULL) {
+			value = g_key_file_get_string(keyfile,
+						provider->identifier,
+						key, NULL);
+			connman_provider_set_string(provider, key, value);
+			g_free(value);
+		}
+		idx += 1;
+	}
+	g_strfreev(settings);
+
+	g_key_file_free(keyfile);
+	return 0;
+}
+
+static int connman_provider_save(struct connman_provider *provider)
+{
+	GKeyFile *keyfile;
+
+	DBG("provider %p", provider);
+
+	keyfile = g_key_file_new();
+	if (keyfile == NULL)
+		return -ENOMEM;
+
+	g_key_file_set_string(keyfile, provider->identifier,
+			"Name", provider->name);
+	g_key_file_set_string(keyfile, provider->identifier,
+			"Type", provider->type);
+	g_key_file_set_string(keyfile, provider->identifier,
+			"Host", provider->host);
+	g_key_file_set_string(keyfile, provider->identifier,
+			"VPN.Domain", provider->domain);
+
+	if (provider->driver != NULL && provider->driver->save != NULL)
+		provider->driver->save(provider, keyfile);
+
+	__connman_storage_save_provider(keyfile, provider->identifier);
+        g_key_file_free(keyfile);
+
+	return 0;
+}
+
 static struct connman_provider *connman_provider_lookup(const char *identifier)
 {
 	struct connman_provider *provider = NULL;
@@ -82,23 +152,6 @@ static struct connman_provider *connman_provider_lookup(const char *identifier)
 	return provider;
 }
 
-struct connman_provider *connman_provider_ref(struct connman_provider *provider)
-{
-	DBG("provider %p", provider);
-
-	if (connman_element_ref(&provider->element) == NULL)
-		return NULL;
-
-	return provider;
-}
-
-void connman_provider_unref(struct connman_provider *provider)
-{
-	DBG("provider %p", provider);
-
-	connman_element_unref(&provider->element);
-}
-
 static gboolean match_driver(struct connman_provider *provider,
 				struct connman_provider_driver *driver)
 {
@@ -137,16 +190,69 @@ static int provider_probe(struct connman_provider *provider)
 	return 0;
 }
 
+static void provider_remove(struct connman_provider *provider)
+{
+	if (provider->driver != NULL) {
+		provider->driver->remove(provider);
+		provider->driver = NULL;
+	}
+}
+
+static int provider_register(struct connman_provider *provider)
+{
+	connman_provider_load(provider);
+	return provider_probe(provider);
+}
+
+static void provider_unregister(struct connman_provider *provider)
+{
+	provider_remove(provider);
+}
+
+struct connman_provider *connman_provider_ref(struct connman_provider *provider)
+{
+	DBG("provider %p refcount %d", provider, provider->refcount + 1);
+
+	__sync_fetch_and_add(&provider->refcount, 1);
+
+	return provider;
+}
+
+static void provider_destruct(struct connman_provider *provider)
+{
+	DBG("provider %p", provider);
+
+	g_free(provider->name);
+	g_free(provider->type);
+	g_free(provider->host);
+	g_free(provider->domain);
+	g_free(provider->identifier);
+	g_hash_table_destroy(provider->routes);
+	g_hash_table_destroy(provider->setting_strings);
+}
+
+void connman_provider_unref(struct connman_provider *provider)
+{
+	DBG("provider %p refcount %d", provider, provider->refcount - 1);
+
+	if (__sync_fetch_and_sub(&provider->refcount, 1) != 1)
+		return;
+
+	provider_remove(provider);
+
+	provider_destruct(provider);
+}
+
 static int provider_indicate_state(struct connman_provider *provider,
 					enum connman_service_state state)
 {
 	DBG("state %d", state);
 
-	__connman_service_indicate_state(provider->vpn_service, state,
+	__connman_service_ipconfig_indicate_state(provider->vpn_service, state,
 					CONNMAN_IPCONFIG_TYPE_IPV4);
 
-	return __connman_service_indicate_state(provider->vpn_service, state,
-					CONNMAN_IPCONFIG_TYPE_IPV6);
+	return __connman_service_ipconfig_indicate_state(provider->vpn_service,
+					state, CONNMAN_IPCONFIG_TYPE_IPV6);
 }
 
 int __connman_provider_disconnect(struct connman_provider *provider)
@@ -218,6 +324,8 @@ int __connman_provider_remove(const char *path)
 
 		if (g_strcmp0(srv_path, path) == 0) {
 			DBG("Removing VPN %s", provider->identifier);
+
+			provider_unregister(provider);
 			g_hash_table_remove(provider_hash,
 						provider->identifier);
 			return 0;
@@ -232,7 +340,7 @@ static void provider_append_routes(gpointer key, gpointer value,
 {
 	struct connman_route *route = value;
 	struct connman_provider *provider = user_data;
-	int index = provider->element.index;
+	int index = provider->index;
 
 	if (route->family == AF_INET6) {
 		unsigned char prefix_len = atoi(route->netmask);
@@ -319,6 +427,8 @@ int connman_provider_indicate_error(struct connman_provider *provider,
 					enum connman_provider_error error)
 {
 	enum connman_service_error service_error;
+	const char *path;
+	int ret;
 
 	switch (error) {
 	case CONNMAN_PROVIDER_ERROR_LOGIN_FAILED:
@@ -335,8 +445,12 @@ int connman_provider_indicate_error(struct connman_provider *provider,
 		break;
 	}
 
-	return __connman_service_indicate_error(provider->vpn_service,
+	ret = __connman_service_indicate_error(provider->vpn_service,
 							service_error);
+	path = __connman_service_get_path(provider->vpn_service);
+	__connman_provider_remove(path);
+
+	return ret;
 }
 
 static void unregister_provider(gpointer data)
@@ -349,23 +463,9 @@ static void unregister_provider(gpointer data)
 	provider->vpn_service = NULL;
 	__connman_service_put(service);
 
-	connman_element_unregister(&provider->element);
 	connman_provider_unref(provider);
 }
 
-static void provider_destruct(struct connman_element *element)
-{
-	struct connman_provider *provider = element->private;
-
-	DBG("provider %p", provider);
-
-	g_free(provider->name);
-	g_free(provider->type);
-	g_free(provider->domain);
-	g_free(provider->identifier);
-	g_hash_table_destroy(provider->routes);
-}
-
 static void destroy_route(gpointer user_data)
 {
 	struct connman_route *route = user_data;
@@ -380,17 +480,15 @@ static void provider_initialize(struct connman_provider *provider)
 {
 	DBG("provider %p", provider);
 
-	__connman_element_initialize(&provider->element);
-
-	provider->element.private = provider;
-	provider->element.destruct = provider_destruct;
-
+	provider->index = 0;
 	provider->name = NULL;
 	provider->type = NULL;
 	provider->domain = NULL;
 	provider->identifier = NULL;
 	provider->routes = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 					NULL, destroy_route);
+	provider->setting_strings = g_hash_table_new_full(g_str_hash, g_str_equal,
+							g_free, g_free);
 }
 
 static struct connman_provider *connman_provider_new(void)
@@ -401,6 +499,8 @@ static struct connman_provider *connman_provider_new(void)
 	if (provider == NULL)
 		return NULL;
 
+	provider->refcount = 1;
+
 	DBG("provider %p", provider);
 	provider_initialize(provider);
 
@@ -425,8 +525,7 @@ static struct connman_provider *connman_provider_get(const char *identifier)
 
 	g_hash_table_insert(provider_hash, provider->identifier, provider);
 
-	provider->element.name = g_strdup(identifier);
-	connman_element_register(&provider->element, NULL);
+	provider->name = g_strdup(identifier);
 
 	return provider;
 }
@@ -450,10 +549,9 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 {
 	struct connman_provider *provider;
 	DBusMessageIter iter, array;
-	const char *type = NULL, *name = NULL, *service_path = NULL;
+	const char *type = NULL, *name = NULL, *service_path;
 	const char *host = NULL, *domain = NULL;
 	char *ident;
-	gboolean created = FALSE;
 	int err;
 
 	dbus_message_iter_init(msg, &iter);
@@ -485,17 +583,13 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 		dbus_message_iter_next(&array);
 	}
 
-	if (host == NULL || domain == NULL) {
-		err = -EINVAL;
-		goto failed;
-	}
+	if (host == NULL || domain == NULL)
+		return -EINVAL;
 
 	DBG("Type %s name %s", type, name);
 
-	if (type == NULL || name == NULL) {
-		err = -EOPNOTSUPP;
-		goto failed;
-	}
+	if (type == NULL || name == NULL)
+		return -EOPNOTSUPP;
 
 	ident = g_strdup_printf("%s_%s", host, domain);
 	provider_dbus_ident(ident);
@@ -503,23 +597,23 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 	DBG("ident %s", ident);
 
 	provider = connman_provider_lookup(ident);
-
 	if (provider == NULL) {
-		created = TRUE;
 		provider = connman_provider_get(ident);
-		if (provider) {
-			provider->host = g_strdup(host);
-			provider->domain = g_strdup(domain);
-			provider->name = g_strdup(name);
-			provider->type = g_strdup(type);
+		if (provider == NULL) {
+			DBG("can not create provider");
+			g_free(ident);
+			return -EOPNOTSUPP;
 		}
+
+		provider->host = g_strdup(host);
+		provider->domain = g_strdup(domain);
+		g_free(provider->name);
+		provider->name = g_strdup(name);
+		provider->type = g_strdup(type);
+
+		provider_register(provider);
 	}
 
-	if (provider == NULL) {
-		DBG("can not create provider");
-		err = -EOPNOTSUPP;
-		goto failed;
-	}
 	dbus_message_iter_init(msg, &iter);
 	dbus_message_iter_recurse(&iter, &array);
 
@@ -545,26 +639,21 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 
 	g_free(ident);
 
-	if (provider == NULL) {
-		err = -EOPNOTSUPP;
-		goto failed;
-	}
-
-	if (created == TRUE)
-		provider_probe(provider);
-
-	if (provider->vpn_service == NULL)
+	if (provider->vpn_service == NULL) {
 		provider->vpn_service =
 			__connman_service_create_from_provider(provider);
-	if (provider->vpn_service == NULL) {
-		err = -EOPNOTSUPP;
-		goto failed;
-	}
+		if (provider->vpn_service == NULL) {
+			err = -EOPNOTSUPP;
+			goto unref;
+		}
 
-	err = __connman_service_connect(provider->vpn_service);
-	if (err < 0 && err != -EINPROGRESS)
-		goto failed;
+		err = __connman_service_connect(provider->vpn_service);
+		if (err < 0 && err != -EINPROGRESS)
+			goto failed;
+	} else
+		DBG("provider already connected");
 
+	connman_provider_save(provider);
 	service_path = __connman_service_get_path(provider->vpn_service);
 	g_dbus_send_reply(connection, msg,
 				DBUS_TYPE_OBJECT_PATH, &service_path,
@@ -572,15 +661,13 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 	return 0;
 
 failed:
-	if (provider != NULL && created == TRUE) {
-		DBG("can not connect delete provider");
-		connman_provider_unref(provider);
+	__connman_service_put(provider->vpn_service);
+	provider->vpn_service = NULL;
 
-		if (provider->vpn_service != NULL) {
-			__connman_service_put(provider->vpn_service);
-			provider->vpn_service = NULL;
-		}
-	}
+unref:
+	DBG("can not connect, delete provider");
+
+	connman_provider_unref(provider);
 
 	return err;
 }
@@ -604,9 +691,16 @@ int connman_provider_set_string(struct connman_provider *provider,
 	} else if (g_str_equal(key, "Name") == TRUE) {
 		g_free(provider->name);
 		provider->name = g_strdup(value);
-	}
-
-	return connman_element_set_string(&provider->element, key, value);
+	} else if (g_str_equal(key, "Host") == TRUE) {
+		g_free(provider->host);
+		provider->host = g_strdup(value);
+	} else if (g_str_equal(key, "VPN.Domain") == TRUE) {
+		g_free(provider->domain);
+		provider->domain = g_strdup(value);
+	} else
+		g_hash_table_replace(provider->setting_strings,
+				g_strdup(key), g_strdup(value));
+	return 0;
 }
 
 const char *connman_provider_get_string(struct connman_provider *provider,
@@ -618,8 +712,12 @@ const char *connman_provider_get_string(struct connman_provider *provider,
 		return provider->type;
 	else if (g_str_equal(key, "Name") == TRUE)
 		return provider->name;
+	else if (g_str_equal(key, "Host") == TRUE)
+		return provider->host;
+	else if (g_str_equal(key, "VPN.Domain") == TRUE)
+		return provider->domain;
 
-	return connman_element_get_string(&provider->element, key);
+	return g_hash_table_lookup(provider->setting_strings, key);
 }
 
 void *connman_provider_get_data(struct connman_provider *provider)
@@ -674,12 +772,12 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 	__connman_ipconfig_set_index(ipconfig, index);
 
 done:
-	provider->element.index = index;
+	provider->index = index;
 }
 
 int connman_provider_get_index(struct connman_provider *provider)
 {
-	return provider->element.index;
+	return provider->index;
 }
 
 int connman_provider_set_ipaddress(struct connman_provider *provider,
@@ -736,12 +834,14 @@ int connman_provider_set_nameservers(struct connman_provider *provider,
 
 	__connman_service_nameserver_clear(provider->vpn_service);
 
-	if (nameservers != NULL)
-		nameservers_array = g_strsplit(nameservers, " ", 0);
+	if (nameservers == NULL)
+		return 0;
+
+	nameservers_array = g_strsplit(nameservers, " ", 0);
 
 	for (i = 0; nameservers_array[i] != NULL; i++) {
 		__connman_service_nameserver_append(provider->vpn_service,
-							nameservers_array[i]);
+						nameservers_array[i], FALSE);
 	}
 
 	g_strfreev(nameservers_array);
@@ -856,9 +956,17 @@ int connman_provider_append_route(struct connman_provider *provider,
 
 const char *connman_provider_get_driver_name(struct connman_provider *provider)
 {
+	if (provider->driver == NULL)
+		return NULL;
+
 	return provider->driver->name;
 }
 
+const char *connman_provider_get_save_group(struct connman_provider *provider)
+{
+	return provider->identifier;
+}
+
 static gint compare_priority(gconstpointer a, gconstpointer b)
 {
 	return 0;
@@ -888,12 +996,12 @@ void connman_provider_driver_unregister(struct connman_provider_driver *driver)
 	driver_list = g_slist_remove(driver_list, driver);
 }
 
-static void provider_remove(gpointer key, gpointer value,
+static void provider_remove_all(gpointer key, gpointer value,
 						gpointer user_data)
 {
 	struct connman_provider *provider = value;
 
-	g_hash_table_remove(provider_hash, provider->identifier);
+	__connman_provider_remove(provider->identifier);
 }
 
 static void provider_offline_mode(connman_bool_t enabled)
@@ -901,7 +1009,7 @@ static void provider_offline_mode(connman_bool_t enabled)
 	DBG("enabled %d", enabled);
 
 	if (enabled == TRUE)
-		g_hash_table_foreach(provider_hash, provider_remove, NULL);
+		g_hash_table_foreach(provider_hash, provider_remove_all, NULL);
 
 }
 
