@@ -23,6 +23,7 @@
 #include <config.h>
 #endif
 
+#include <errno.h>
 #include <string.h>
 
 #include <gdbus.h>
@@ -31,10 +32,10 @@
 
 static DBusConnection *connection;
 
-static GHashTable *rfkill_table;
-static GHashTable *device_table;
 static GSList *technology_list = NULL;
 
+static connman_bool_t global_offlinemode;
+
 struct connman_rfkill {
 	unsigned int index;
 	enum connman_service_type type;
@@ -45,29 +46,31 @@ struct connman_rfkill {
 enum connman_technology_state {
 	CONNMAN_TECHNOLOGY_STATE_UNKNOWN   = 0,
 	CONNMAN_TECHNOLOGY_STATE_OFFLINE   = 1,
-	CONNMAN_TECHNOLOGY_STATE_AVAILABLE = 2,
-	CONNMAN_TECHNOLOGY_STATE_BLOCKED   = 3,
-	CONNMAN_TECHNOLOGY_STATE_ENABLED   = 4,
-	CONNMAN_TECHNOLOGY_STATE_CONNECTED = 5,
+	CONNMAN_TECHNOLOGY_STATE_ENABLED   = 2,
+	CONNMAN_TECHNOLOGY_STATE_CONNECTED = 3,
 };
 
 struct connman_technology {
-	gint refcount;
+	int refcount;
 	enum connman_service_type type;
 	enum connman_technology_state state;
 	char *path;
 	GHashTable *rfkill_list;
 	GSList *device_list;
-	gint enabled;
-	gint blocked;
+	int enabled;
 	char *regdom;
 
 	connman_bool_t tethering;
 	char *tethering_ident;
 	char *tethering_passphrase;
 
+	connman_bool_t enable_persistent; /* Save the tech state */
+
 	struct connman_technology_driver *driver;
 	void *driver_data;
+
+	DBusMessage *pending_reply;
+	guint pending_timeout;
 };
 
 static GSList *driver_list = NULL;
@@ -151,6 +154,8 @@ static void tethering_changed(struct connman_technology *technology)
 void connman_technology_tethering_notify(struct connman_technology *technology,
 							connman_bool_t enabled)
 {
+	GSList *list;
+
 	DBG("technology %p enabled %u", technology, enabled);
 
 	if (technology->tethering == enabled)
@@ -162,14 +167,21 @@ void connman_technology_tethering_notify(struct connman_technology *technology,
 
 	if (enabled == TRUE)
 		__connman_tethering_set_enabled();
-	else
-		__connman_tethering_set_disabled();
+	else {
+		for (list = technology_list; list; list = list->next) {
+			struct connman_technology *other_tech = list->data;
+			if (other_tech->tethering == TRUE)
+				break;
+		}
+		if (list == NULL)
+			__connman_tethering_set_disabled();
+	}
 }
 
 static int set_tethering(struct connman_technology *technology,
-				const char *bridge, connman_bool_t enabled)
+				connman_bool_t enabled)
 {
-	const char *ident, *passphrase;
+	const char *ident, *passphrase, *bridge;
 
 	ident = technology->tethering_ident;
 	passphrase = technology->tethering_passphrase;
@@ -178,6 +190,10 @@ static int set_tethering(struct connman_technology *technology,
 			technology->driver->set_tethering == NULL)
 		return -EOPNOTSUPP;
 
+	bridge = __connman_tethering_get_bridge();
+	if (bridge == NULL)
+		return -EOPNOTSUPP;
+
 	if (technology->type == CONNMAN_SERVICE_TYPE_WIFI &&
 	    (ident == NULL || passphrase == NULL))
 		return -EINVAL;
@@ -253,10 +269,6 @@ static const char *state2string(enum connman_technology_state state)
 		break;
 	case CONNMAN_TECHNOLOGY_STATE_OFFLINE:
 		return "offline";
-	case CONNMAN_TECHNOLOGY_STATE_AVAILABLE:
-		return "available";
-	case CONNMAN_TECHNOLOGY_STATE_BLOCKED:
-		return "blocked";
 	case CONNMAN_TECHNOLOGY_STATE_ENABLED:
 		return "enabled";
 	case CONNMAN_TECHNOLOGY_STATE_CONNECTED:
@@ -303,6 +315,115 @@ static const char *get_name(enum connman_service_type type)
 	return NULL;
 }
 
+static void load_state(struct connman_technology *technology)
+{
+	GKeyFile *keyfile;
+	gchar *identifier;
+	GError *error = NULL;
+	connman_bool_t enable;
+
+	DBG("technology %p", technology);
+
+	keyfile = __connman_storage_load_global();
+	/* Fallback on disabling technology if file not found. */
+	if (keyfile == NULL) {
+		technology->enable_persistent = FALSE;
+		return;
+	}
+
+	identifier = g_strdup_printf("%s", get_name(technology->type));
+	if (identifier == NULL)
+		goto done;
+
+	enable = g_key_file_get_boolean(keyfile, identifier, "Enable", &error);
+	if (error == NULL)
+		technology->enable_persistent = enable;
+	else {
+		technology->enable_persistent = FALSE;
+		g_clear_error(&error);
+	}
+done:
+	g_free(identifier);
+
+	g_key_file_free(keyfile);
+
+	return;
+}
+
+static void save_state(struct connman_technology *technology)
+{
+	GKeyFile *keyfile;
+	gchar *identifier;
+
+	DBG("technology %p", technology);
+
+	keyfile = __connman_storage_load_global();
+	if (keyfile == NULL)
+		keyfile = g_key_file_new();
+
+	identifier = g_strdup_printf("%s", get_name(technology->type));
+	if (identifier == NULL)
+		goto done;
+
+	g_key_file_set_boolean(keyfile, identifier, "Enable",
+				technology->enable_persistent);
+
+done:
+	g_free(identifier);
+
+	__connman_storage_save_global(keyfile);
+
+	g_key_file_free(keyfile);
+
+	return;
+}
+
+connman_bool_t __connman_technology_get_offlinemode(void)
+{
+	return global_offlinemode;
+}
+
+static void connman_technology_save_offlinemode()
+{
+	GKeyFile *keyfile;
+
+	keyfile = __connman_storage_load_global();
+	if (keyfile == NULL)
+		keyfile = g_key_file_new();
+
+	g_key_file_set_boolean(keyfile, "global",
+					"OfflineMode", global_offlinemode);
+
+	__connman_storage_save_global(keyfile);
+
+	g_key_file_free(keyfile);
+
+	return;
+}
+
+static connman_bool_t connman_technology_load_offlinemode()
+{
+	GKeyFile *keyfile;
+	GError *error = NULL;
+	connman_bool_t offlinemode;
+
+	/* If there is a error, we enable offlinemode */
+	keyfile = __connman_storage_load_global();
+	if (keyfile == NULL)
+		return TRUE;
+
+	offlinemode = g_key_file_get_boolean(keyfile, "global",
+						"OfflineMode", &error);
+	if (error != NULL) {
+		offlinemode = TRUE;
+		g_clear_error(&error);
+	}
+
+	g_key_file_free(keyfile);
+
+	return offlinemode;
+}
+
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *message, void *user_data)
 {
@@ -377,7 +498,6 @@ static DBusMessage *set_property(DBusConnection *conn,
 	if (g_str_equal(name, "Tethering") == TRUE) {
 		int err;
 		connman_bool_t tethering;
-		const char *bridge;
 
 		if (type != DBUS_TYPE_BOOLEAN)
 			return __connman_error_invalid_arguments(msg);
@@ -387,9 +507,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (technology->tethering == tethering)
 			return __connman_error_in_progress(msg);
 
-		bridge = __connman_tethering_get_bridge();
-
-		err = set_tethering(technology, bridge, tethering);
+		err = set_tethering(technology, tethering);
 		if (err < 0)
 			return __connman_error_failed(msg, -err);
 
@@ -457,7 +575,7 @@ static struct connman_technology *technology_get(enum connman_service_type type)
 
 	technology = technology_find(type);
 	if (technology != NULL) {
-		g_atomic_int_inc(&technology->refcount);
+		__sync_fetch_and_add(&technology->refcount, 1);
 		goto done;
 	}
 
@@ -479,8 +597,11 @@ static struct connman_technology *technology_get(enum connman_service_type type)
 							NULL, free_rfkill);
 	technology->device_list = NULL;
 
+	technology->pending_reply = NULL;
 	technology->state = CONNMAN_TECHNOLOGY_STATE_OFFLINE;
 
+	load_state(technology);
+
 	if (g_dbus_register_interface(connection, technology->path,
 					CONNMAN_TECHNOLOGY_INTERFACE,
 					technology_methods, technology_signals,
@@ -521,7 +642,7 @@ static void technology_put(struct connman_technology *technology)
 {
 	DBG("technology %p", technology);
 
-	if (g_atomic_int_dec_and_test(&technology->refcount) == FALSE)
+	if (__sync_fetch_and_sub(&technology->refcount, 1) != 1)
 		return;
 
 	if (technology->driver) {
@@ -547,7 +668,7 @@ static void technology_put(struct connman_technology *technology)
 void __connman_technology_add_interface(enum connman_service_type type,
 				int index, const char *name, const char *ident)
 {
-	GSList *list;
+	struct connman_technology *technology;
 
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
@@ -567,27 +688,20 @@ void __connman_technology_add_interface(enum connman_service_type type,
 	connman_info("Create interface %s [ %s ]", name,
 				__connman_service_type2string(type));
 
-	technology_get(type);
+	technology = technology_get(type);
 
-	for (list = technology_list; list; list = list->next) {
-		struct connman_technology *technology = list->data;
+	if (technology == NULL || technology->driver == NULL
+			|| technology->driver->add_interface == NULL)
+		return;
 
-		if (technology->type != type)
-			continue;
-
-		if (technology->driver == NULL)
-			continue;
-
-		if (technology->driver->add_interface)
-			technology->driver->add_interface(technology,
-							index, name, ident);
-	}
+	technology->driver->add_interface(technology,
+					index, name, ident);
 }
 
 void __connman_technology_remove_interface(enum connman_service_type type,
 				int index, const char *name, const char *ident)
 {
-	GSList *list;
+	struct connman_technology *technology;
 
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
@@ -607,25 +721,13 @@ void __connman_technology_remove_interface(enum connman_service_type type,
 	connman_info("Remove interface %s [ %s ]", name,
 				__connman_service_type2string(type));
 
-	for (list = technology_list; list; list = list->next) {
-		struct connman_technology *technology = list->data;
+	technology = technology_find(type);
 
-		if (technology->type != type)
-			continue;
+	if (technology == NULL || technology->driver == NULL)
+		return;
 
-		if (technology->driver == NULL)
-			continue;
-
-		if (technology->driver->remove_interface)
-			technology->driver->remove_interface(technology, index);
-
-		technology_put(technology);
-	}
-}
-
-static void unregister_technology(gpointer data)
-{
-	struct connman_technology *technology = data;
+	if (technology->driver->remove_interface)
+		technology->driver->remove_interface(technology, index);
 
 	technology_put(technology);
 }
@@ -644,16 +746,11 @@ int __connman_technology_add_device(struct connman_device *device)
 	if (technology == NULL)
 		return -ENXIO;
 
-	g_hash_table_insert(device_table, device, technology);
-
-	if (g_atomic_int_get(&technology->blocked))
-		goto done;
-
-	technology->state = CONNMAN_TECHNOLOGY_STATE_AVAILABLE;
-
-	state_changed(technology);
-
-done:
+	if (technology->enable_persistent && !global_offlinemode)
+		__connman_device_enable(device);
+	/* if technology persistent state is offline */
+	if (!technology->enable_persistent)
+		__connman_device_disable(device);
 
 	technology->device_list = g_slist_append(technology->device_list,
 								device);
@@ -671,7 +768,7 @@ int __connman_technology_remove_device(struct connman_device *device)
 	type = __connman_device_get_service_type(device);
 	__connman_notifier_unregister(type);
 
-	technology = g_hash_table_lookup(device_table, device);
+	technology = technology_find(type);
 	if (technology == NULL)
 		return -ENXIO;
 
@@ -682,79 +779,254 @@ int __connman_technology_remove_device(struct connman_device *device)
 		state_changed(technology);
 	}
 
-	g_hash_table_remove(device_table, device);
-
 	return 0;
 }
 
-int __connman_technology_enable_device(struct connman_device *device)
+static gboolean technology_pending_reply(gpointer user_data)
+{
+	struct connman_technology *technology = user_data;
+	DBusMessage *reply;
+
+	/* Power request timedout, send ETIMEDOUT. */
+	if (technology->pending_reply != NULL) {
+		reply = __connman_error_failed(technology->pending_reply, ETIMEDOUT);
+		if (reply != NULL)
+			g_dbus_send_message(connection, reply);
+
+		dbus_message_unref(technology->pending_reply);
+		technology->pending_reply = NULL;
+		technology->pending_timeout = 0;
+	}
+
+	return FALSE;
+}
+
+int __connman_technology_enabled(enum connman_service_type type)
 {
 	struct connman_technology *technology;
-	enum connman_service_type type;
 
-	DBG("device %p", device);
-
-	technology = g_hash_table_lookup(device_table, device);
+	technology = technology_find(type);
 	if (technology == NULL)
 		return -ENXIO;
 
-	if (g_atomic_int_get(&technology->blocked))
-		return -ERFKILL;
-
-	type = __connman_device_get_service_type(device);
-	__connman_notifier_enable(type);
-
-	if (g_atomic_int_exchange_and_add(&technology->enabled, 1) == 0) {
+	if (__sync_fetch_and_add(&technology->enabled, 1) == 0) {
+		__connman_notifier_enable(type);
 		technology->state = CONNMAN_TECHNOLOGY_STATE_ENABLED;
 		state_changed(technology);
 	}
 
+	if (technology->pending_reply != NULL) {
+		g_dbus_send_reply(connection, technology->pending_reply, DBUS_TYPE_INVALID);
+		dbus_message_unref(technology->pending_reply);
+		g_source_remove(technology->pending_timeout);
+		technology->pending_reply = NULL;
+		technology->pending_timeout = 0;
+	}
+
 	return 0;
 }
 
-int __connman_technology_disable_device(struct connman_device *device)
+int __connman_technology_enable(enum connman_service_type type, DBusMessage *msg)
 {
 	struct connman_technology *technology;
-	enum connman_service_type type;
 	GSList *list;
+	int err = 0;
+	int ret = -ENODEV;
+	DBusMessage *reply;
 
-	DBG("device %p", device);
+	DBG("type %d enable", type);
 
-	type = __connman_device_get_service_type(device);
-	__connman_notifier_disable(type);
-
-	technology = g_hash_table_lookup(device_table, device);
-	if (technology == NULL)
-		return -ENXIO;
-
-	if (g_atomic_int_dec_and_test(&technology->enabled) == TRUE) {
-		technology->state = CONNMAN_TECHNOLOGY_STATE_AVAILABLE;
-		state_changed(technology);
+	technology = technology_find(type);
+	if (technology == NULL) {
+		err = -ENXIO;
+		goto done;
 	}
 
+	if (technology->pending_reply != NULL) {
+		err = -EBUSY;
+		goto done;
+	}
+
+	if (msg != NULL) {
+		/*
+		 * This is a bit of a trick. When msg is not NULL it means
+		 * thats technology_enable was invoked from the manager API. Hence we save
+		 * the state here.
+		 */
+		technology->enable_persistent = TRUE;
+		save_state(technology);
+	}
+
+	__connman_rfkill_block(technology->type, FALSE);
+
+	/*
+	 * An empty device list means that devices in the technology
+	 * were rfkill blocked. The unblock above will enable the devs.
+	 */
+	if (technology->device_list == NULL)
+		return 0;
+
 	for (list = technology->device_list; list; list = list->next) {
 		struct connman_device *device = list->data;
 
-		if (__connman_device_get_blocked(device) == FALSE)
-			return 0;
+		err = __connman_device_enable(device);
+		/*
+		 * err = 0 : Device was enabled right away.
+		 * If atleast one device gets enabled, we consider
+		 * the technology to be enabled.
+		 */
+		if (err == 0)
+			ret = 0;
 	}
 
-	technology->state = CONNMAN_TECHNOLOGY_STATE_BLOCKED;
+done:
+	if (ret == 0) {
+		if (msg != NULL)
+			g_dbus_send_reply(connection, msg, DBUS_TYPE_INVALID);
+		return ret;
+	}
+
+	if (msg != NULL) {
+		if (err == -EINPROGRESS) {
+			technology->pending_reply = dbus_message_ref(msg);
+			technology->pending_timeout = g_timeout_add_seconds(10,
+					technology_pending_reply, technology);
+		} else {
+			reply = __connman_error_failed(msg, -err);
+			if (reply != NULL)
+				g_dbus_send_message(connection, reply);
+		}
+	}
+
+	return err;
+}
+
+int __connman_technology_disabled(enum connman_service_type type)
+{
+	struct connman_technology *technology;
+
+	technology = technology_find(type);
+	if (technology == NULL)
+		return -ENXIO;
+
+	if (technology->pending_reply != NULL) {
+		g_dbus_send_reply(connection, technology->pending_reply, DBUS_TYPE_INVALID);
+		dbus_message_unref(technology->pending_reply);
+		g_source_remove(technology->pending_timeout);
+		technology->pending_reply = NULL;
+		technology->pending_timeout = 0;
+	}
+
+	if (__sync_fetch_and_sub(&technology->enabled, 1) != 1)
+		return 0;
+
+	__connman_notifier_disable(type);
+	technology->state = CONNMAN_TECHNOLOGY_STATE_OFFLINE;
 	state_changed(technology);
 
 	return 0;
 }
 
-static void technology_blocked(struct connman_technology *technology,
-				connman_bool_t blocked)
+int __connman_technology_disable(enum connman_service_type type, DBusMessage *msg)
 {
+	struct connman_technology *technology;
 	GSList *list;
+	int err = 0;
+	int ret = -ENODEV;
+	DBusMessage *reply;
+
+	DBG("type %d disable", type);
+
+	technology = technology_find(type);
+	if (technology == NULL) {
+		err = -ENXIO;
+		goto done;
+	}
+
+	if (technology->pending_reply != NULL) {
+		err = -EBUSY;
+		goto done;
+	}
+
+	if (technology->tethering == TRUE)
+		set_tethering(technology, FALSE);
+
+	if (msg != NULL) {
+		technology->enable_persistent = FALSE;
+		save_state(technology);
+	}
+
+	__connman_rfkill_block(technology->type, TRUE);
 
 	for (list = technology->device_list; list; list = list->next) {
 		struct connman_device *device = list->data;
 
-		__connman_device_set_blocked(device, blocked);
+		err = __connman_device_disable(device);
+		if (err == 0)
+			ret = 0;
 	}
+
+done:
+	if (ret == 0) {
+		if (msg != NULL)
+			g_dbus_send_reply(connection, msg, DBUS_TYPE_INVALID);
+		return ret;
+	}
+
+	if (msg != NULL) {
+		if (err == -EINPROGRESS) {
+			technology->pending_reply = dbus_message_ref(msg);
+			technology->pending_timeout = g_timeout_add_seconds(10,
+					technology_pending_reply, technology);
+		} else {
+			reply = __connman_error_failed(msg, -err);
+			if (reply != NULL)
+				g_dbus_send_message(connection, reply);
+		}
+	}
+
+	return err;
+}
+
+int __connman_technology_set_offlinemode(connman_bool_t offlinemode)
+{
+	GSList *list;
+	int err = -EINVAL;
+
+	if (global_offlinemode == offlinemode)
+		return 0;
+
+	DBG("offlinemode %s", offlinemode ? "On" : "Off");
+
+	/*
+	 * This is a bit tricky. When you set offlinemode, there is no
+	 * way to differentiate between attempting offline mode and
+	 * resuming offlinemode from last saved profile. We need that
+	 * information in rfkill_update, otherwise it falls back on the
+	 * technology's persistent state. Hence we set the offline mode here
+	 * but save it & call the notifier only if its successful.
+	 */
+
+	global_offlinemode = offlinemode;
+
+	/* Traverse technology list, enable/disable each technology. */
+	for (list = technology_list; list; list = list->next) {
+		struct connman_technology *technology = list->data;
+
+		if (offlinemode)
+			err = __connman_technology_disable(technology->type, NULL);
+
+		if (!offlinemode && technology->enable_persistent)
+			err = __connman_technology_enable(technology->type, NULL);
+	}
+
+	if (err == 0 || err == -EINPROGRESS || err == -EALREADY) {
+		connman_technology_save_offlinemode();
+		__connman_notifier_offlinemode(offlinemode);
+	} else
+		global_offlinemode = connman_technology_load_offlinemode();
+
+	return err;
 }
 
 int __connman_technology_add_rfkill(unsigned int index,
@@ -764,7 +1036,6 @@ int __connman_technology_add_rfkill(unsigned int index,
 {
 	struct connman_technology *technology;
 	struct connman_rfkill *rfkill;
-	connman_bool_t blocked;
 
 	DBG("index %u type %d soft %u hard %u", index, type,
 							softblock, hardblock);
@@ -777,40 +1048,49 @@ int __connman_technology_add_rfkill(unsigned int index,
 	if (rfkill == NULL)
 		return -ENOMEM;
 
+	__connman_notifier_register(type);
+
 	rfkill->index = index;
 	rfkill->type = type;
 	rfkill->softblock = softblock;
 	rfkill->hardblock = hardblock;
 
-	g_hash_table_replace(rfkill_table, &rfkill->index, technology);
-
 	g_hash_table_replace(technology->rfkill_list, &rfkill->index, rfkill);
 
-	blocked = (softblock || hardblock) ? TRUE : FALSE;
-	if (blocked == FALSE)
+	if (hardblock) {
+		DBG("%s is switched off.", get_name(type));
 		return 0;
+	}
 
-	if (g_atomic_int_exchange_and_add(&technology->blocked, 1) == 0) {
-		technology_blocked(technology, TRUE);
-
-		technology->state = CONNMAN_TECHNOLOGY_STATE_BLOCKED;
-		state_changed(technology);
+	/*
+	 * If Offline mode is on, we softblock the device if it isnt already.
+	 * If Offline mode is off, we rely on the persistent state of tech.
+	 */
+	if (global_offlinemode) {
+		if (!softblock)
+			return __connman_rfkill_block(type, TRUE);
+	} else {
+		if (technology->enable_persistent && softblock)
+			return __connman_rfkill_block(type, FALSE);
+		/* if technology persistent state is offline */
+		if (!technology->enable_persistent && !softblock)
+			return __connman_rfkill_block(type, TRUE);
 	}
 
 	return 0;
 }
 
 int __connman_technology_update_rfkill(unsigned int index,
+					enum connman_service_type type,
 						connman_bool_t softblock,
 						connman_bool_t hardblock)
 {
 	struct connman_technology *technology;
 	struct connman_rfkill *rfkill;
-	connman_bool_t blocked, old_blocked;
 
 	DBG("index %u soft %u hard %u", index, softblock, hardblock);
 
-	technology = g_hash_table_lookup(rfkill_table, &index);
+	technology = technology_find(type);
 	if (technology == NULL)
 		return -ENXIO;
 
@@ -818,47 +1098,37 @@ int __connman_technology_update_rfkill(unsigned int index,
 	if (rfkill == NULL)
 		return -ENXIO;
 
-	old_blocked = (rfkill->softblock || rfkill->hardblock) ? TRUE : FALSE;
-	blocked = (softblock || hardblock) ? TRUE : FALSE;
+	if (rfkill->softblock == softblock &&
+		rfkill->hardblock == hardblock)
+		return 0;
 
 	rfkill->softblock = softblock;
 	rfkill->hardblock = hardblock;
 
-	if (blocked == old_blocked)
+	if (hardblock) {
+		DBG("%s is switched off.", get_name(type));
 		return 0;
+	}
 
-	if (blocked) {
-		guint n_blocked;
-
-		n_blocked =
-			g_atomic_int_exchange_and_add(&technology->blocked, 1);
-		if (n_blocked != g_hash_table_size(technology->rfkill_list) - 1)
-			return 0;
-
-		technology_blocked(technology, blocked);
-		technology->state = CONNMAN_TECHNOLOGY_STATE_BLOCKED;
-		state_changed(technology);
-	} else {
-		if (g_atomic_int_dec_and_test(&technology->blocked) == FALSE)
-			return 0;
-
-		technology_blocked(technology, blocked);
-		technology->state = CONNMAN_TECHNOLOGY_STATE_AVAILABLE;
-		state_changed(technology);
+	if (!global_offlinemode) {
+		if (technology->enable_persistent && softblock)
+			return __connman_rfkill_block(type, FALSE);
+		if (!technology->enable_persistent && !softblock)
+			return __connman_rfkill_block(type, TRUE);
 	}
 
 	return 0;
 }
 
-int __connman_technology_remove_rfkill(unsigned int index)
+int __connman_technology_remove_rfkill(unsigned int index,
+					enum connman_service_type type)
 {
 	struct connman_technology *technology;
 	struct connman_rfkill *rfkill;
-	connman_bool_t blocked;
 
 	DBG("index %u", index);
 
-	technology = g_hash_table_lookup(rfkill_table, &index);
+	technology = technology_find(type);
 	if (technology == NULL)
 		return -ENXIO;
 
@@ -866,46 +1136,20 @@ int __connman_technology_remove_rfkill(unsigned int index)
 	if (rfkill == NULL)
 		return -ENXIO;
 
-	blocked = (rfkill->softblock || rfkill->hardblock) ? TRUE : FALSE;
-
 	g_hash_table_remove(technology->rfkill_list, &index);
 
-	g_hash_table_remove(rfkill_table, &index);
-
-	if (blocked &&
-		g_atomic_int_dec_and_test(&technology->blocked) == TRUE) {
-		technology_blocked(technology, FALSE);
-		technology->state = CONNMAN_TECHNOLOGY_STATE_AVAILABLE;
-		state_changed(technology);
-	}
+	technology_put(technology);
 
 	return 0;
 }
 
-connman_bool_t __connman_technology_get_blocked(enum connman_service_type type)
-{
-	struct connman_technology *technology;
-
-	technology = technology_find(type);
-	if (technology == NULL)
-		return FALSE;
-
-	if (g_atomic_int_get(&technology->blocked))
-		return TRUE;
-
-	return FALSE;
-}
-
 int __connman_technology_init(void)
 {
 	DBG("");
 
 	connection = connman_dbus_get_connection();
 
-	rfkill_table = g_hash_table_new_full(g_int_hash, g_int_equal,
-						NULL, unregister_technology);
-	device_table = g_hash_table_new_full(g_direct_hash, g_direct_equal,
-						NULL, unregister_technology);
+	global_offlinemode = connman_technology_load_offlinemode();
 
 	return 0;
 }
@@ -914,8 +1158,5 @@ void __connman_technology_cleanup(void)
 {
 	DBG("");
 
-	g_hash_table_destroy(device_table);
-	g_hash_table_destroy(rfkill_table);
-
 	dbus_connection_unref(connection);
 }
